#include "problem_gen.hpp"

#define VERSION "1.0"

#include <vector>
#include <iostream>
#include <random>
#include <algorithm>

auto generate( ProblemGenerator& pg ) -> void {
    int customerAmount, foodAmount, t;
    std::vector<int> orders{ };

    std::cout << "Welcome to the Qck-bite problem generator!" << std::endl;
    std::cout << "Please, specify the amount of customers: " << std::endl;
    std::cin >> customerAmount;

    if ( customerAmount <= 0 ) {
        std::cerr << "ERROR: The customer amount cannot be equal or lower than zero." << std::endl;
        std::exit( 1 );
    }

    std::cout << "And, now, the amount of different foods: " << std::endl;
    std::cin >> foodAmount;

    if ( foodAmount <= 0 ) {
        std::cerr << "ERROR: The food amount cannot be equal or lower than zero." << std::endl;
        std::exit( 1 );
    }

    std::cout << "Lastly, specify the orders of each customer (type -1 for random): " << std::endl;
    std::cin >> t;

    int i{ 0 };
    while ( t != -1 ) {
        if ( t >= foodAmount ) {
            std::cerr << "ERROR: The user specified an invalid order." << std::endl;
            std::exit( 1 );
        }

        orders.push_back( t );

        if ( i++ >= customerAmount - 1 ) {
            break;
        }

        std::cin >> t;
    }

    if ( i < customerAmount && i > 0 ) {
        std::cerr << "ERROR: The user did not specify the orders of every customer." << std::endl;
        std::exit( 1 );
    }

    pg
            .beginProblem( "qck-bite" )
            .beginObjects()
            .addObject( "waiter", "waiter" )
            .addObject( "cook", "cook" )
            .addObject( "bench", "bench" )
            .addObject( "stove", "stove" );

    for ( int j = 0; j < foodAmount; j++ ) {
        std::stringstream label;
        label << "food" << j;

        pg.addObject( label.str( ), "food" );
    }

    for ( int j = 0; j < customerAmount; j++ ) {
        std::stringstream labelOrder, labelCustomer;
        labelOrder << "order" << j;
        labelCustomer << "customer" << j;

        pg.addObject(labelOrder.str( ), "order" );
        pg.addObject(labelCustomer.str( ), "client" );
    }

    // Check if we should randomize the orders.
    if ( orders.empty( ) ) {
        int n = 0;

        // God, C++ lambdas are sexy...
        const auto gen = [&] {
            n = ( n + 1 ) % customerAmount;
            return n;
        };

        // Populate the orders vector with all possible foods.
        std::generate_n( std::back_inserter( orders ),
                         customerAmount,
                         gen );

        // Shuffle the vector to make it random.
        std::shuffle( orders.begin( ),
                     orders.end( ),
                     std::mt19937{ std::random_device{ }( ) } );
    }

    pg
        .endObjects()
        .beginHtn();

    for ( int j = 0; j < customerAmount; j++ ) {
        std::stringstream task, labelClient, labelFood;
        labelClient << "client" << j;
        labelFood   << "food"   << orders[j];

        task << "dothedeed";
        task << " bench";
        task << " " << labelClient.str( );
        task << " cook";
        task << " " << labelFood.str( );
        task << " stove";
        task << " waiter";

        // Add the task.
        pg.addSubtask( task.str( ) );
    }

    pg
            .endHtn()
            .beginInit();

    for ( int j = 0; j < customerAmount; j++ ) {
        std::stringstream labelOrder, labelCustomer, labelFood;

        labelCustomer << "client" << j;
        labelOrder    << "order"  << j;
        labelFood     << "food"   << orders[j];

        pg.addPredicate( "client_has_order",
                         { labelCustomer.str( ), labelOrder.str( ) } );
        pg.addPredicate( "order_refers_to",
                         { labelOrder.str( ), labelFood.str( ) } );
    }

    pg
            .endInit()
            .endProblem()
            .writeAll();

    std::cout << "Successfully written the problem to problem_generated.hddl" << std::endl;
}

auto help( ) -> void {
    std::cout << "Qck-bite problem generator\n"
                 "Created by Caio Urnau, @aprxl.\n\n"
                 "This is a wizard for the creation of HDDL problems for the Qck-bite domain. In order to use\n"
                 "it, just follow the instructions presented to you upon executing the program.\n\n"
                 "List of valid arguments:\n"
                 "  --help | -h:    Displays this message.\n"
                 "  --version | -v: Shows the software version.";
}

auto version( ) -> void {
    std::cout << "Qck-bite problem generator\n"
                 "Created by Caio Urnau, @aprxl.\n"
                 "Version " << VERSION << std::endl;
}

auto invalid( ) -> void {
    std::cout << "Qck-bite problem generator\n"
                 "Created by Caio Urnau, @aprxl.\n"
                 "The specified argument is invalid. Use the '--help' argument to see all valid arguments.\n";
}

auto main( int argc, char** argv ) -> int {
    // Check if the user specified any arguments.
    if ( argc > 1 ) {
        std::vector<std::string> args( argv + 1, argv + argc );

        if ( args[0] == "--help" || args[0] == "-h" )
            help( );

        else if ( args[0] == "--version" || args[0] == "-v" )
            version( );

        else
            invalid( );

        return 0;
    }

    ProblemGenerator pg( "problem_generated.txt" );

    if ( !pg.isOpen( ) ) {
        std::cerr << "ERROR: The target file could not be opened." << std::endl;
        return 1;
    }

    generate( pg );
	return 0;
}
