#pragma once

#ifndef PROBLEM_GEN_HH
#define PROBLEM_GEN_HH
#include <fstream>
#include <initializer_list>
#include <sstream>
#include <string>

/// A class responsible for writing HDDL problems. Serves no bigger purpose than
/// being a helper class, as most data still needs to be specified by the user.
class ProblemGenerator {
private:
  int m_Depth;
  std::stringstream m_Buffer;
  std::ofstream m_File;

public:
  explicit ProblemGenerator() = delete;
  explicit ProblemGenerator(const std::string &path) : m_Buffer{}, m_Depth{0} {
    m_File = std::ofstream(path);

    // The file isn't open, thus isn't working.
    if (!m_File.is_open()) {
      std::fprintf(stderr, "The output file cannot be opened. (%s)",
                   path.c_str());
    }
  }

  ~ProblemGenerator() {
    // Make sure to close the file before going out of scope.
    m_File.close();
  }

  [[nodiscard]] bool isOpen() noexcept { return m_File.is_open(); }

  [[nodiscard]] std::ofstream &getFile() noexcept { return m_File; }

  bool setPath(const std::string &path) {
    // Close the file and open a new one.
    m_File.close();
    m_File = std::ofstream(path);

    // Check if file was successfully opened.
    if (!m_File.is_open()) {
      std::fprintf(stderr, "The output file cannot be opened. (%s)",
                   path.c_str());

      return false;
    }

    return true;
  }

  ProblemGenerator &
  beginProblem(const std::string &domainName,
               const std::string &problemName = "generated") noexcept {
    // Write to our buffer the beginning of our problem.
    m_Buffer << "(define" << std::endl;
    m_Buffer << "  (problem " << problemName << ")" << std::endl;
    m_Buffer << "  (:domain " << domainName << ")" << std::endl;
    m_Depth++;

    return *this;
  }

  ProblemGenerator &endProblem() noexcept {
    // Cute method.
    m_Buffer << ")";
    m_Depth--;

    return *this;
  }

  ProblemGenerator &beginObjects() noexcept {
    m_Buffer << "   (:objects" << std::endl;
    m_Depth++;

    return *this;
  }

  ProblemGenerator &addObject(const std::string &label,
                              const std::string &type = "object") noexcept {

    m_Buffer << "       " << label << " - " << type << std::endl;

    return *this;
  }

  ProblemGenerator &endObjects() noexcept {
    m_Buffer << "   )" << std::endl;
    m_Depth--;

    return *this;
  }

  ProblemGenerator &beginInit() noexcept {
    m_Buffer << "   (:init" << std::endl;
    m_Depth++;

    return *this;
  }

  ProblemGenerator &
  addPredicate(const std::string &predicate,
               std::initializer_list<std::string> args) noexcept {

    m_Buffer << "       ( ";
    m_Buffer << predicate << " ";

    for (auto &arg : args) {
      m_Buffer << "?" << arg << " ";
    }

    m_Buffer << ")" << std::endl;

    return *this;
  }

  ProblemGenerator &endInit() noexcept {
    m_Buffer << "   )" << std::endl;
    m_Depth--;

    return *this;
  }

  ProblemGenerator &beginHtn() noexcept {
    m_Buffer << "   (:htn" << std::endl;
    m_Buffer << "       :parameters ()" << std::endl;
    m_Buffer << "       :subtasks (and" << std::endl;
    m_Depth += 2;

    return *this;
  }

  ProblemGenerator &addSubtask(const std::string &task) noexcept {
    m_Buffer << "           (" << task << ")" << std::endl;

    return *this;
  }

  ProblemGenerator &endHtn() noexcept {
    m_Buffer << "       )" << std::endl;
    m_Buffer << "   )" << std::endl;
    m_Depth -= 2;

    return *this;
  }

  bool writeAll() noexcept {
    if (!isOpen()) {
      std::fprintf(stderr, "Unable to write buffer to file, file is closed.\n");
      return false;
    }

    if (m_Depth != 0) {
      std::fprintf(
          stderr,
          "Unable to write buffer to file, depth is not zero. (There are "
          "parentheses that haven't been properly closed, depth: %d)\n",
          m_Depth);
      return false;
    }

    m_File << m_Buffer.str();
    return true;
  }
};
#endif
